# Good Colors:
#
#     LIGHTRED_EX
#     CYAN | LIGHTCYAN_EX
#     LIGHTBLUE_EX 
#     LIGHTGREEN_EX
#     YELLOW 
#     MAGENTA

#
# SEARCH     : print_\(\w\+\)_line(\(.*\))
# SUBSTITUTE : s/print_\(\w\+\)_line(\(.*\))/print(Code.\U\1 + \2)
#
# Doesn’t work for nested function calls example: 
#
#   offsets = [ make_string_green(offset) if offset == None else str(offset) for offset in offsets ]
#
# SEARCH     : make_string_\(\w\+\)(\(.*\))
# SUBSTITUTE : s/make_string_\(\w\+\)(\(.*\))/Code.\U\1\E+\2

from .timer import Timer
timer = Timer()

import os
import sys


class Maker:
    RESET = '\x1b[39m'
    def __init__(self, code):
        self.COLOR = code
    def __add__(self, item):
        str_item = str(item)
        color_str_item = self.COLOR + str_item + self.RESET
        return color_str_item

class Code:
    BLACK           = Maker('\x1b[30m')
    RED             = Maker('\x1b[31m')
    GREEN           = Maker('\x1b[32m')
    YELLOW          = Maker('\x1b[33m')
    BLUE            = Maker('\x1b[34m')
    MAGENTA         = Maker('\x1b[35m')
    CYAN            = Maker('\x1b[36m')
    WHITE           = Maker('\x1b[37m')
    LIGHTBLACK_EX   = Maker('\x1b[90m')
    LIGHTRED_EX     = Maker('\x1b[91m')
    LIGHTGREEN_EX   = Maker('\x1b[92m')
    LIGHTYELLOW_EX  = Maker('\x1b[93m')
    LIGHTBLUE_EX    = Maker('\x1b[94m')
    LIGHTMAGENTA_EX = Maker('\x1b[95m')
    LIGHTCYAN_EX    = Maker('\x1b[96m')
    LIGHTWHITE_EX   = Maker('\x1b[97m')

def color_filepath(filepath):
    import os.path
    dirname = Code.LIGHTBLACK_EX + (os.path.dirname(filepath) + os.path.sep)
    basename = Code.LIGHTCYAN_EX + os.path.basename(filepath)
    result = dirname + basename 
    return result


def print_samples():
    value = [1, 2,3, 4]

    print()
    print(f"{'RED':16} This is the way {Code.RED + value}, I hope you understand")
    print(f"{'LIGHTRED_EX':16} This is the way {Code.LIGHTRED_EX + value}, I hope you understand")
    print()

    print(f"{'GREEN':16} This is the way {Code.GREEN + value}, I hope you understand")
    print(f"{'LIGHTGREEN_EX':16} This is the way {Code.LIGHTGREEN_EX + value}, I hope you understand")
    print()

    print(f"{'YELLOW':16} This is the way {Code.YELLOW + value}, I hope you understand")
    print(f"{'LIGHTYELLOW_EX':16} This is the way {Code.LIGHTYELLOW_EX + value}, I hope you understand")
    print()

    print(f"{'BLUE':16} This is the way {Code.BLUE + value}, I hope you understand")
    print(f"{'LIGHTBLUE_EX':16} This is the way {Code.LIGHTBLUE_EX + value}, I hope you understand")
    print()

    print(f"{'BLACK':16} This is the way {Code.BLACK + value}, I hope you understand")
    print(f"{'LIGHTBLACK_EX':16} This is the way {Code.LIGHTBLACK_EX + value}, I hope you understand")
    print()

    print(f"{'MAGENTA':16} This is the way {Code.MAGENTA + value}, I hope you understand")
    print(f"{'LIGHTMAGENTA_EX':16} This is the way {Code.LIGHTMAGENTA_EX + value}, I hope you understand")
    print()

    print(f"{'CYAN':16} This is the way {Code.CYAN + value}, I hope you understand")
    print(f"{'LIGHTCYAN_EX':16} This is the way {Code.LIGHTCYAN_EX + value}, I hope you understand")
    print()

    print(f"{'LIGHTWHITE_EX':16} This is the way {Code.LIGHTWHITE_EX + value}, I hope you understand")
    print(f"{'WHITE':16} This is the way {Code.WHITE + value}, I hope you understand")
    print()

def supports_color():
    # from django
    """
    Return True if the running system's terminal supports color,
    and False otherwise.
    """

    def vt_codes_enabled_in_windows_registry():
        """
        Check the Windows Registry to see if VT code handling has been enabled
        by default, see https://superuser.com/a/1300251/447564.
        """
        try:
            # winreg is only available on Windows.
            import winreg
        except ImportError:
            return False
        else:
            try:
                reg_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, "Console")
                reg_key_value, _ = winreg.QueryValueEx(reg_key, "VirtualTerminalLevel")
            except FileNotFoundError:
                return False
            else:
                return reg_key_value == 1

    # isatty is not always implemented, #6223.
    is_a_tty = hasattr(sys.stdout, "isatty") and sys.stdout.isatty()

    return is_a_tty and (
        sys.platform != "win32"
        or "ANSICON" in os.environ
        or
        # Windows Terminal supports VT codes.
        "WT_SESSION" in os.environ
        or
        # Microsoft Visual Studio Code's built-in terminal supports colors.
        os.environ.get("TERM_PROGRAM") == "vscode"
        or vt_codes_enabled_in_windows_registry()
    )

def enable_cmd_color_windows_10_1607():
    # Black magic!
    # https://stackoverflow.com/a/39675059

    # To enable: https://superuser.com/questions/413073/windows-console-with-ansi-colors-handling
    # Go to regedit
    #
    # dword is 32-bit
    # [HKEY_CURRENT_USER\Console]
    # "VirtualTerminalLevel"=dword:00000001
    #
    return
    if supports_color() is False:
        os.system('') #enable VT100 Escape Sequence for WINDOWS 10 Ver. 1607
        timer.start("import ctypes")
        from ctypes import windll
        timer.print("import ctypes")
        windll.kernel32.SetConsoleMode(windll.kernel32.GetStdHandle(-11), 7)

enable_cmd_color_windows_10_1607()

if __name__ == "__main__":
    print_samples()




# https://www.python.org/dev/peps/pep-0508/#environment-markers
from setuptools import setup
import sys
def get_new_version():
    import urllib.request
    import json
    package_name = "waivek"
    url = f"https://pypi.org/pypi/{package_name}/json"
    req = urllib.request.urlopen(url)
    D = json.load(req)
    version = D["info"]["version"]
    major, minor, patch = version.split(".")
    patch = int(patch) + 1
    new_version = f"{major}.{minor}.{patch}"
    print(f"{package_name} version: {version} -> {new_version}")
    return new_version

def remove_dist_directory():
    import shutil
    shutil.rmtree("dist", ignore_errors=True)

if len(sys.argv) >=3 and sys.argv[1] == 'sdist' and sys.argv[2] == 'bdist_wheel':
    remove_dist_directory()
    new_version = get_new_version()
else:
    print("Usage: python setup.py sdist bdist_wheel")
    sys.exit(1)

py_modules = ['color', 'common', 'db', 'error', 'frame', 'get', 'ic', 'print_utils', 'reltools', 'timer', 'trace']
py_modules = ['waivek.' + x for x in py_modules]
long_description = open('README.md').read()
setup(
        name='waivek',
        version=new_version,
        packages=['waivek'],
        py_modules = py_modules,
        install_requires = [
            'aiohttp==3.8.1',
            'executing==0.8.2',
            'python_dateutil==2.8.2',
            'timeago==1.0.14',
            'pysqlite3-binary; platform_system=="Linux"'
        ],
        long_description=long_description,
        long_description_content_type='text/markdown'
)





# SETTRACE vs. SETPROFILE
# =======================
#
# https://stackoverflow.com/a/59088868
#
# sys.settrace() only reports on Python code, not built-in callables or those defined in compiled extensions
#
# When using sys.settrace() returning a function object instead of None lets you trace other events within the frame,
# this is the 'local' trace function. You can re-use the same function object for this. You can disable per-line events
# by setting `frame.f_trace_lines = False`.


"""


    INPUT  : arguments are file-storable: text, float, int, list(json), datetime(int)
    OUTPUT : returns are file-storable and can be compared for correctness when re-run with same inputs
    EPOCH  : time the function was called / run

Optional, Return value changes w.r.t. __file__ of caller
Exampe - rel2abs(string), ic()

    FILE   : File that called this function
    FUNC   : Function that called this function

"""

from .timer import Timer   # Single Use
timer = Timer()
from .color import Code    # Multi-Use
from .error import handler # Single Use
from .ic import ic, ib
ib; ic; Code

from datetime import datetime
import sys
import pickle
from time import time
import os.path

gf = None
path_function_pairs = []
trace_dictionaries = []
total_time = 0

def path_to_module_name(path):
    frame_index = 1
    while True:
        frame = sys._getframe(frame_index)
        if frame.f_code.co_filename == path:
            module_name = frame.f_globals['__name__']
            return module_name
        frame_index = frame_index + 1

def trace(function_names):
    if function_names == []:
        return
    global path_function_pairs
    frame = sys._getframe(1)
    filepath = frame.f_code.co_filename
    for function_name in function_names:
        pair = (filepath, function_name)
        path_function_pairs.append(pair)
    path_function_pairs = list(set(path_function_pairs))
    import threading
    threading.settrace(handle_trace_event)
    sys.settrace(handle_trace_event)

def rerun():
    global done
    done = True
    for trace_D in trace_dictionaries:
        result = pickle.loads(trace_D['result_pkl'])
        kwargs = pickle.loads(trace_D['kwargs_pkl'])
        function_name = trace_D['function_name']
        filepath = trace_D['filepath']
        function = get_function(filepath, function_name)
        calc_result = function(**kwargs)
        function_string = f"{function_name}(**{kwargs})"
        cmp_string = f"{function_string} == {repr(result)}"
        if calc_result == result:
            print(Code.GREEN + "✓", cmp_string)
        else:
             print(Code.RED + "✗", cmp_string)




def target(name: str, age: int, dob: datetime, face_bytes: bytes, items: list[str], maps: dict[str, str], known='Value'):
    global x
    # ic(locals())
    return "1, 2, 3"


def to_table(D):
    from .ic import Table
    table = Table()
    for row in D.items():
        table.row(row)
    import textwrap
    table_string = "\n".join(str(table).split("\n")[1:])
    table_string = textwrap.dedent(table_string)
    return table_string

def greeting():
    print("Hello, World!")

def get_function(path, function_name):
    from importlib import import_module
    # import os.path
    # module_name,_ = os.path.splitext(os.path.basename(path))
    module_name = path_to_module_name(path)
    module = import_module(module_name, path)
    function = getattr(module, function_name)
    return function

def add(x, y):
    # print(f"Sum Is: {x+y}")
    return x+y


def call_and_return_tracer(frame, event, arg, events = []):
    global total_time
    start = time()
    print("handle_trace_event", event, os.path.basename(frame.f_code.co_filename), frame.f_code.co_name)
    if event == 'call':
        print(f"Entering: {frame.f_code.co_name}")
        frame.f_trace_lines = False
        events.append(event)
        print(f"Entering: (event): {events!r}")
        total_time = total_time + time() - start
        return call_and_return_tracer
    elif event == 'return':
        events.append(event)
        print(f"Returning: {arg!r}")
        print(f"Returning (event): {events!r}")
        total_time = total_time + time() - start

def handle_trace_event(frame, event, arg, function_D = {}):
    # print("handle_trace_event", event, os.path.basename(frame.f_code.co_filename), frame.f_code.co_name)
    global total_time
    global gf
    global path_function_pairs
    start = time()
    gf = frame
    if event.startswith("c_"):
        return

    pair = (frame.f_code.co_filename, frame.f_code.co_name)
    if pair not in path_function_pairs:
        return
   
    # done = False
    # if done:
    #     return

    frame_hash = hex(hash(frame))

    function_D["function_name"] = frame.f_code.co_name
    function_D["filepath"] = frame.f_code.co_filename
    if event == 'call':
        function_D["start_epoch"] = time()
        function_D["end_epoch"] = None
        function_D["kwargs_pkl"] = pickle.dumps(frame.f_locals)
        frame.f_trace_lines = False
        total_time = total_time + time() - start
        return handle_trace_event
    elif event == 'return':
        function_D["end_epoch"] = time()
        function_D["result_pkl"] = pickle.dumps(arg)

    if event == 'return':
        trace_dictionaries.append(function_D)
        total_time = total_time + time() - start

def main():
    trace(['add'])
    # sys.settrace(call_and_return_tracer)
    # foo(spam, "universe")
    # for i in range(10):
    #     add(1, 2)
    import threading
    thread = threading.Thread(target=add, args=[3, 4])
    thread.start()
    thread.join()
    ic(trace_dictionaries)



if __name__ == "__main__":
    main(); exit()
    with handler():
        main()



import sys; sys.path = [ "C:/users/vivek/Documents/Python/" ] + sys.path
from .timer import Timer   # Single Use
timer = Timer()
import ast
import linecache
from .color import Code    # Multi-Use
from .error import handler # Single Use
from .ic import ic, ib     # Multi-Use, import time: 70ms - 110ms

from types import FrameType

path2lines: dict[str, list[str]] = {}

class Frame:
    def __init__(self, frame: FrameType):
        self.path = frame.f_code.co_filename
        self.lineno = frame.f_lineno
        self.line = getline(self.path, self.lineno)
        self.raw = frame
        # frame.f_back, frame.f_locals, frame.f_globals

def getline(path: str, lineno: int):
    global path2lines
    if path not in path2lines:
        with open(path, "rb") as f:
            lines = f.read().decode("utf-8").split("\n")
            path2lines[path] = lines
    index = lineno-1
    lines = path2lines[path]
    return lines[index]

def inspect_dot_stack():
    frames = [ sys._getframe(1) ]
    while frame := frames[-1].f_back:
        frames.append(frame)
    return [ Frame(frame) for frame in frames ]

def frame_gen():
    frame = sys._getframe(1)
    while frame:
        yield frame
        frame = frame.f_back


# IMPLEMENTATION 1:
# =================
def to_dict(*variables):
    current_function_name = sys._getframe(0).f_code.co_name
    frame = sys._getframe(1)
    line = linecache.getline(frame.f_code.co_filename, frame.f_lineno).strip()
    # inside_string = line.split(current_function_name)[1][1:-1]
    inside_string = line.replace(current_function_name, "")[1:-1]
    names = [ name.strip() for name in inside_string.split(",") ]
    D = { name: frame.f_locals[name] for name in names }
    return D

# IMPLEMENTATION 2:
# =================
def create_dict(*args):
    frame = sys._getframe(1)
    node = ast.parse(linecache.getline(frame.f_code.co_filename, frame.f_lineno).strip()).body[0].value
    return { arg.id: frame.f_locals[arg.id] for arg in node.args }



def repr_exists(obj):
    string = repr(obj)
    if string[0] != '<' or string[-1] != '>':
        return True
    import re
    m = re.match(r'\<.* object at 0x[0-9A-F]+\>', string)
    if m is None:
        return True
    else:
        return False
    # '<color.Maker object at 0x00000220322BB5E0>'

def unused_silencer():
    Code, ic, ib

def call_node_to_string(node, local_D):
    for index, arg in enumerate(node.args):
        match type(arg):
            case ast.Name:
                value = local_D[arg.id]
                if repr_exists(value):
                    node.args[index] = ast.Constant(local_D[arg.id])
                else:
                    A = repr(value.__class__.__name__)
                    C = repr(value.__dict__)
                    value_repr = f"type({A}, ({value.__class__.__name__},), {C})"
                    value_call = ast.parse(value_repr).body[0].value
                    node.args[index] = value_call

            case ast.Call:
                node.args[index] = call_node_to_string(arg, local_D)

    # print(ast.unparse(node))
    return node


def frame_pretty():

    caller = sys._getframe(2)

    caller_path = caller.f_code.co_filename
    caller_lineno = caller.f_lineno
    caller_line = linecache.getline(caller_path, caller_lineno).strip()
    caller_node = ast.parse(caller_line).body[0].value
    caller_locals = caller.f_locals

    # ic(D)
    caller_line
    eval_line = ast.unparse(call_node_to_string(caller_node, caller_locals))
    D = {
        "lineno": caller_lineno,
        "path"  : caller_path,
        "line": caller_line,
        "eval_line": eval_line
    }
    ic(D)




def bar(z):
    return 'Blast'

class Point:
    def __init__(self, item):
        from random import randint
        self.item = item
        self.value = randint(0, 100)
    def ten():
        return 10

# foo    (x, y, 1, 'Sup', bar(z), [1, 2, 3, 4, 5], now, pt)
def foo(x, y, u,     z, result,               L,  dt, pt, string='Hello'):
    frames = sys._current_frames()
    frame_pretty()
    print(pt.value)
    return x + y

def main():

    ic(getline(__file__, 6))
    return

    from datetime import datetime
    import re
    result = re.split('a', 'bab')
    datetime.now()
    x = 1
    y = 2
    z = 3
    pt = Point(1)
    now = datetime.now()
    # ic(repr_exists(1))
    # ic(repr_exists(now))
    # ic(repr_exists(pt))
    L = [1, 2, 3, 4, 5]
    foo(x, y, 1, 'Sup', bar(z), [1, 2, 3, 4, 5], now, pt)

    # create_dict(x, y)

if __name__ == "__main__":
    with handler():
        main()
